""" ifrit utility implementations using lxml.objectify

Note:  nodes returned from this module can *not* have marker interfaces
       added to them at the instance level, due to restricttions in the
       base class (extension type) required by objectify.

$Id: objectify_utils.py,v 1.4 2007/02/17 18:26:16 tseaver Exp $
"""
from lxml.etree import XMLParser
from lxml.etree import parse as _parse
from lxml.objectify import XML as _XML

from zope.interface import directlyProvides
from zope.interface import implements

from ifrit.interfaces import IStringParser
from ifrit.interfaces import IStreamParser

_parser = XMLParser(remove_blank_text=True)

def fromstring(xmlstring):
    return _XML(xmlstring, _parser)

directlyProvides(fromstring, IStringParser)

def parse(xmlstream):
    return _parse(xmlstream, _parser)

directlyProvides(parse, IStreamParser)



""" ifrit package interfaces

$Id: interfaces.py,v 1.2 2007/02/09 21:41:42 tseaver Exp $
"""
from zope.interface import Interface

class IStringParser(Interface):
    """ Utility interface for a parser which can be called passing a string.

    o Examples:  elementtree.ElementTree.XML
    """
    def __call__(xmlstring):
        """ Parse 'xmlstring', and return the root ElementTree node.
        """

class IStreamParser(Interface):
    """ Utility interface for a parser which can be called passing a stream.

    o Examples:  elementtree.ElementTree.parse
    """
    def __call__(xmlstream):
        """ Parse 'xmlstream', and return the root ElementTree node.
        """

class IElement(Interface):
    """ Adapter interface.
    """

class IElementFactory(Interface):
    """ Adapter factory interface.
    """
    def __call__(context):
        """ Adapt context to an element interface
        
        o Adapted object will provide an interface derived from IElement.
        """

class IXMLSerialization(Interface):
    """ Utility interface.
    """
    def serialize(buffer):
        """ Write an XML representation of 'context' into 'buffer'.
        """



""" ifrit utility implementations using lxml

"""
from lxml.etree import XML as _XML
from lxml.etree import parse as _parse
from lxml.etree import XMLParser

from zope.interface import directlyProvides

from ifrit.interfaces import IStringParser
from ifrit.interfaces import IStreamParser

_parser = XMLParser()

def fromstring(xmlstring):
    return _XML(xmlstring, _parser)

directlyProvides(fromstring, IStringParser)

def parse(xmlstream):
    return _parse(xmlstream, _parser)

directlyProvides(parse, IStreamParser)



import unittest
from StringIO import StringIO

from zope.interface import Interface
from zope.interface import implements
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.app.testing.placelesssetup import PlacelessSetup
from zope.app.testing import ztapi

class OFSFileAsXMLTests(unittest.TestCase):

    def _getTargetClass(self):
        from ifrit.adapters import OFSFileAsXML
        return OFSFileAsXML

    def _makeOne(self, data):
        file = DummyFile(data)
        return self._getTargetClass()(file)

    def test_context_with_string_data(self):
        from ifrit.interfaces import IElement
        _STRING_PROFILE = """<?xml version="1.0"?><document/>"""
        adapted = self._makeOne(_STRING_PROFILE)
        buffer = StringIO()
        adapted.serialize(buffer)
        self.assertEqual(buffer.getvalue(), _STRING_PROFILE)

    def test_context_with_unicode_data(self):
        from ifrit.interfaces import IElement
        _UNICODE_PROFILE = u"""<?xml version="1.0"?><document/>"""
        adapted = self._makeOne(_UNICODE_PROFILE)
        buffer = StringIO()
        adapted.serialize(buffer)
        self.assertEqual(buffer.getvalue(), _UNICODE_PROFILE.encode('utf8'))

    def test_context_with_pdata_data(self):
        from ifrit.interfaces import IElement
        _PDATA_PROFILE = DummyPData('<?xml version="1.0"?>',
                                    DummyPData('<document/>'))
        adapted = self._makeOne(_PDATA_PROFILE)
        buffer = StringIO()
        adapted.serialize(buffer)
        self.assertEqual(buffer.getvalue(), '<?xml version="1.0"?><document/>')


class ElementMakerTests(unittest.TestCase, PlacelessSetup):

    def setUp(self):
        PlacelessSetup.setUp(self)

    def tearDown(self):
        PlacelessSetup.tearDown(self)

    def _makeOne(self, path=None, parser_name=None, multiple=False):
        from zope.interface import directlyProvides
        from ifrit.adapters import ElementMaker
        file = DummyFile()
        maker = ElementMaker(path, parser_name=parser_name, multiple=multiple)
        return maker(file)

    def _fakeSerializer(self, data):
        from ifrit.interfaces import IXMLSerialization
        class DummyAdapter:
            def __init__(self, context):
                pass
            def serialize(self, buffer):
                buffer.write(data)
        ztapi.provideAdapter(IDummy, IXMLSerialization, DummyAdapter, '')

    def _registerStreamParsers(self):
        from elementtree.ElementTree import parse as et_parse
        from ifrit.interfaces import IStreamParser
        from ifrit.lxml_utils import parse as lx_parse
        from ifrit.objectify_utils import parse as ob_parse
        ztapi.provideUtility(IStreamParser, et_parse, 'elementtree')
        ztapi.provideUtility(IStreamParser, lx_parse, 'lxml')
        ztapi.provideUtility(IStreamParser, ob_parse, 'objectify')

    def test_with_no_path_returns_root(self):
        from zope.interface.verify import verifyObject
        _PROFILE = '<?xml version="1.0"?><document><section/></document>'
        self._fakeSerializer(_PROFILE)
        self._registerStreamParsers()

        for parser_name in ('elementtree', 'lxml', 'objectify'):
            adapted = self._makeOne(path=None, parser_name=parser_name)
            self.assertEqual(adapted.tag, 'document')
            self.assertEqual(adapted.find('./section').tag, 'section')

    def test_with_path_returns_child(self):
        from zope.interface.verify import verifyObject
        _PROFILE = '<?xml version="1.0"?><document><section/></document>'
        self._fakeSerializer(_PROFILE)
        self._registerStreamParsers()
        adapted = self._makeOne(path='./section')
        self.assertEqual(adapted.tag, 'section')

    def test_with_invalid_path_returns_None(self):
        _PROFILE = '<?xml version="1.0"?><document><section/></document>'
        self._fakeSerializer(_PROFILE)
        self._registerStreamParsers()
        adapted = self._makeOne(path='./nonesuch')
        self.assertEqual(adapted, None)

    def test_with_path_and_multiple_returns_children(self):
        from zope.interface.verify import verifyObject
        _PROFILE = ('<?xml version="1.0"?>'
                    '<document>'
                     '<section/>'
                     '<section/>'
                    '</document>'
                   )
        self._fakeSerializer(_PROFILE)
        self._registerStreamParsers()
        adapted = self._makeOne(path='./section', multiple=True)
        self.assertEqual(len(adapted), 2)
        for node in adapted:
            self.assertEqual(node.tag, 'section')

class ElementTraversalAdapterTests(unittest.TestCase, PlacelessSetup):

    def setUp(self):
        PlacelessSetup.setUp(self)

    def tearDown(self):
        PlacelessSetup.tearDown(self)

    def _getTargetClass(self):
        from ifrit.adapters import ElementTraversalAdapter
        return ElementTraversalAdapter

    def _makeOne(self):
        from zope.interface import directlyProvides
        file = DummyFile()
        directlyProvides(file, IDummy)
        return self._getTargetClass()(file)

    def test_class_conforms_to__IPublishTraverse(self):
        from zope.interface.verify import verifyClass
        from zope.publisher.interfaces import IPublishTraverse
        verifyClass(IPublishTraverse, self._getTargetClass())

    def test_object_conforms_to_IPublishTraverse(self):
        from zope.interface.verify import verifyObject
        from zope.publisher.interfaces import IPublishTraverse
        from ifrit.interfaces import IElement
        verifyObject(IPublishTraverse, self._makeOne())

    def _makeElementAdapter_etree(self, context):
        from elementtree.ElementTree import XML
        return XML('<dummy><sub/></dummy>')

    def _makeElementAdapter_lxml(self, context):
        from ifrit.lxml_utils import fromstring
        return fromstring('<dummy><sub/></dummy>')

    def _makeElementAdapter_objectify(self, context):
        from ifrit.lxml_utils import fromstring
        return fromstring('<dummy><sub/></dummy>')

    def test_publishTraverse_with_good_name_returns_wrapped_element(self):
        from zope.interface import providedBy, directlyProvides
        from zope.interface.verify import verifyObject
        from zope.publisher.interfaces import IPublishTraverse
        from ifrit.interfaces import IElement

        class IMyElement(IElement):
            pass

        for name, adapter in (('objectify', self._makeElementAdapter_objectify),
                              ('lxml', self._makeElementAdapter_lxml),
                              ('etree', self._makeElementAdapter_etree),
                              ):
            ztapi.provideAdapter(IDummy, IElement, adapter, name)
            adapter = self._makeOne()
            request = {}

            node = adapter.publishTraverse(request, name)

            self.assertEqual(node.tag, 'dummy')
            self.assertEqual(node.__parent__, adapter.context)
            self.assertEqual(node.__name__, name)
            self.assertEqual(list(providedBy(node)), [])

            directlyProvides(node, IMyElement)
            self.failUnless(IMyElement.providedBy(node))

    def test_publishTraverse_with_bad_name_raises_NotFound(self):
        from zope.publisher.interfaces import NotFound
        from ifrit.interfaces import IElement

        ztapi.provideAdapter(IDummy, IElement,
                             self._makeElementAdapter_etree, 'name')
        adapter = self._makeOne()
        request = {}

        self.assertRaises(NotFound, adapter.publishTraverse, request, 'bad')

    def test_publishTraverse_falls_back_to_default(self):
        from zope.component import adapts
        from zope.component import provideAdapter
        from zope.publisher.interfaces.browser import IBrowserView
        from ifrit.interfaces import IElement

        class DummyView(object):
            implements(IBrowserView)
            adapts(IDummy, IBrowserRequest)
            def __init__(self, context, request):
                pass

        provideAdapter(DummyView, name='view.html')
        ztapi.provideAdapter(IDummy, IElement,
                             self._makeElementAdapter_etree, 'name')
        adapter = self._makeOne()
        request = DummyRequest()

        found = adapter.publishTraverse(request, 'view.html')

        self.failUnless(isinstance(found, DummyView))

class ElementTraversalAdapterFactoryTests(unittest.TestCase, PlacelessSetup):

    def setUp(self):
        PlacelessSetup.setUp(self)

    def tearDown(self):
        PlacelessSetup.tearDown(self)

    def _getTargetClass(self):
        from ifrit.adapters import ElementTraversalAdapterFactory
        return ElementTraversalAdapterFactory

    def _makeOne(self, markers):
        return self._getTargetClass()(markers)

    def _makeElementAdapter(self, context):
        #from elementtree.ElementTree import XML
        from ifrit.lxml_utils import fromstring
        return fromstring('<dummy><sub/></dummy>')

    def test___call__with_known_name_returns_marked_element(self):
        from ifrit.interfaces import IElement
        class IMarker(Interface):
            pass
        _MARKERS = {'known': IMarker}

        ztapi.provideAdapter(IDummy, IElement,
                            self._makeElementAdapter, 'known')

        factory = self._makeOne(_MARKERS)
        context = DummyFile()
        request = DummyRequest()
        adapter = factory(context, request)

        node = adapter.publishTraverse(request, 'known')

        self.failUnless(IMarker.providedBy(node))

    def test___call__with_unknown_name_returns_unmarked_element(self):
        from zope.interface import providedBy
        from ifrit.interfaces import IElement
        class IMarker(Interface):
            pass
        _MARKERS = {'known': IMarker}

        ztapi.provideAdapter(IDummy, IElement,
                            self._makeElementAdapter, 'other')

        factory = self._makeOne(_MARKERS)
        context = DummyFile()
        request = DummyRequest()
        adapter = factory(context, request)

        node = adapter.publishTraverse(request, 'other')

        self.assertEquals(len(list(providedBy(node))), 0)

class IDummy(Interface):
    pass

class DummyFile:
    implements(IDummy)
    def __init__(self, data=''):
        self.data = data

class DummyPData:
    def __init__(self, data, next=None):
        self.data = data
        self.next = next

class DummyRequest(object):
    implements(IBrowserRequest)

def test_suite():
    from zope.testing.doctest import DocFileTest
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(OFSFileAsXMLTests))
    suite.addTest(unittest.makeSuite(ElementMakerTests))
    suite.addTest(unittest.makeSuite(ElementTraversalAdapterTests))
    suite.addTest(unittest.makeSuite(ElementTraversalAdapterFactoryTests))
    suite.addTest(DocFileTest('README.txt', package='ifrit'))
    suite.addTest(DocFileTest('nodes_with_markers.txt', package='ifrit'))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')




""" ifrist adapters

$Id: adapters.py,v 1.7 2007/02/17 19:48:15 tseaver Exp $
"""
from StringIO import StringIO

from zope.component import getUtility
from zope.component import queryAdapter
from zope.component import queryMultiAdapter
from zope.interface import implements
from zope.interface import directlyProvides
from zope.interface import Provides
from zope.interface.declarations import getObjectSpecification
from zope.location import LocationProxy
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces import NotFound

from ifrit.interfaces import IElement
from ifrit.interfaces import IElementFactory
from ifrit.interfaces import IStreamParser
from ifrit.interfaces import IXMLSerialization

class OFSFileAsXML(object):
    """ "Convert" OFS.Image.File-like object to XML.

    o Assume that the 'data' element is a string, unicode, or a pdata chunk,
      containing the actual XML to be returned.
    """
    implements(IXMLSerialization)

    def __init__(self, context):
        self.context = context

    def serialize(self, buffer):
        data = self.context.data
        if isinstance(data, str):
            buffer.write(data)
        elif isinstance(data, unicode):
            buffer.write(data.encode('utf8'))
        else:
            while data is not None:
                buffer.write(data.data)
                data = data.next

class ElementMaker(object):
    """ Adapter factories which use paths to find nodes.
    """
    implements(IElementFactory)

    def __init__(self,
                 path=None,
                 parser_name=None,
                 multiple=False,
                ):
        self.path = path
        self.parser_name = parser_name or 'elementtree'
        self.multiple = multiple

    def __call__(self, context):
        buffer = StringIO()

        adapter = IXMLSerialization(context)
        adapter.serialize(buffer)
        buffer.seek(0)

        parser = getUtility(IStreamParser, name=self.parser_name)
        tree = parser(buffer).getroot()

        if self.multiple:
            if self.path is not None:
                nodes = tree.findall(self.path)
            else:
                nodes = [tree.getroot()]

            return nodes

        else:
            if self.path is not None:
                node = tree.find(self.path)
            else:
                node = tree

            return node

class _ProvidedByDescriptor(object):
    def __get__(self, inst, cls=None):
        if inst is None:
            return getObjectSpecification(cls)
        else:
            return getObjectSpecification(inst)

    def __set__(self, inst, value):
        raise TypeError("Can't set __providedBy__ on a decorated object")

class _ElementProxy(LocationProxy):
    __slots__ = ('__name__', '__parent__', '__provides__')
    __providedBy__ = _ProvidedByDescriptor() # defeat parent's version
    def __init__(self, ob, container=None, name=None, *provides):
        super(_ElementProxy, self).__init__(ob, container, name)
        self.__provides__ = Provides(provides)

class ElementTraversalAdapter(object):
    implements(IPublishTraverse)

    def __init__(self, context, markers=None):
        if markers is None:
            markers = {}
        self.context = context
        self.markers = markers

    def publishTraverse(self, request, name):
        """ See IPublishTraverse.
        """
        element = queryAdapter(self.context, IElement, name)
        if element is None:

            view = queryMultiAdapter((self.context, request), name=name)
            if view is not None:
                return view

            raise NotFound(self.context, name, request)

        proxy = _ElementProxy(element, container=self.context, name=name)
        marker = self.markers.get(name)
        if marker is not None:
            directlyProvides(proxy, marker)

        return proxy

class ElementTraversalAdapterFactory(object):

    def __init__(self, markers):
        self.markers = markers

    def __call__(self, context, request=None):
        # We can play as single adapter or view, because request is going
        # to be passed in explicitly to 'publishTraverse'.
        return ElementTraversalAdapter(context, self.markers)



"""Setup for ifrit package

$Id: setup.py,v 1.4 2007/02/17 19:52:23 tseaver Exp $
"""

try:
    from setuptools import setup
except ImportError, e:
    from distutils.core import setup

setup(name='ifrit',
      version='0.3',
      url='http://agendaless.com/Members/tseaver/software/ifrit',
      license='ZPL 2.1',
      description='XPath / ElementTree path adapters',
      author='Tres Seaver, Agendaless Consulting, Inc.',
      author_email='tseaver@agendaless.com',
      long_description='Define adapter factories based on XML paths',
      platform='Any',
      packages=['ifrit', 'ifrit.tests'],
      package_data = {'': ['*.txt', '*.zcml']},
      package_dir = {'': 'src'},
      install_requires=['zope.component',
                        'zope.configuration',
                        'zope.i18nmessageid',
                        'zope.interface',
                        'zope.schema',
                       ],
      extras_require={'elementtree': ['elementtree'],
                      'lxml': ['lxml'],
                      'objectify': ['objectify'],
                     },
      tests_require=['zope.app.testing',
                     'zope.testing.cleanup',
                     'zope.testing.doctest',
                    ],
      include_package_data = True,
      zip_safe = False,
      )




""" ifrit ZCML directives

$Id: zcml.py,v 1.6 2007/02/17 19:48:15 tseaver Exp $
"""
from zope.component import provideAdapter
from zope.interface import Interface
from zope.schema import TextLine
from zope.configuration.fields import GlobalInterface
from zope.configuration.fields import GlobalObject
from zope.configuration.fields import Path
from zope.configuration.fields import PythonIdentifier
from zope.configuration.fields import Tokens
from zope.configuration.fields import Bool
from zope.publisher.interfaces import IPublishTraverse
from zope.publisher.interfaces import IRequest

from ifrit.interfaces import IElement
from ifrit.adapters import ElementMaker
from ifrit.adapters import ElementTraversalAdapterFactory

class IPathAdapter(Interface):
    """ Interface for directive which creates a path-based adapter.

    E.g.:

     <ifrit:path_adapter
       for="package.interfaces.IContent"
       name="adapterName"
       path="some_lxml_path"
       parser_name="objectify"
       module="package.module"
       factory_name="factoryName"
       />

    o Create an adapter factory which uses 'path' to extract a node
      from the serialized XML, as parsed by the utility identifed
      by 'parser_name'.

    o Bind that factory into 'module' under 'factory_name'.

    o Register that factory as providing an 'IElement' adapter for
      the 'for' interface, using the 'name' if provided (otherwise,
      registers it as the default adapter).
    """
    module = GlobalObject(
        title=u"Target module",
        description=u"Module into which the generated adapter factory "
                    u"will be added.",
        required=True,
        )

    factory_name = PythonIdentifier(
        title=u"Factory name",
        description=u"Name of the generated adapter factory.",
        required=True,
        )

    path = TextLine(
        title=u"Path",
        description=u"XPath or ElementTree path expression",
        required=True,
        )

    parser_name = PythonIdentifier(
        title=u"Parser name",
        description=u"Name of IStreamParser utility used to parse XML",
        required=False,
        default=u'elementtree'
        )

    for_ = Tokens(
        title=u"Specifications to be adapted",
        description=u"This should be a list of interfaces or classes",
        value_type=GlobalObject(missing_value=object()),
        required=False,
        )

    name = PythonIdentifier(
        title=u"Adapter name",
        description=u"Name under which the adapter is registered.",
        required=False,
        )

    multiple = Bool(
        title=u"Multiple",
        description=u"Does the factory return a sequence?",
        required=False,
        default=False,
        )


def createPathAdapter(module, factory_name, path, parser_name,
                      for_, name, multiple):
    """ Create an adapter factory using 'path'.
 
    o Seat the new schema into 'module' under 'factory_name'.
    """
    factory = ElementMaker(path, parser_name, multiple)
    setattr(module, factory_name, factory)
    provideAdapter(factory, for_, IElement, name)

def PathAdapterDirective(_context,
                         module,
                         factory_name,
                         path,
                         parser_name='elementtree',
                         for_=None,
                         name='',
                         multiple=False,
                        ):
    # Directive handler for <ifrit:path_adapter> directive.
    _context.action(
        discriminator = (module, factory_name,),
        callable = createPathAdapter,
        args = (module,
                factory_name,
                path,
                parser_name,
                for_,
                name,
                multiple,
               ),
        )

class ITraversalAdapter(Interface):
    """ Interface for directive which creates a traveral adapter factory.

    E.g.:

     <ifrit:traversal_adapter
        for="ifrit.tests.IDummy"
        module="ifrit.tests"
        factory_name="traversalFactory">
      <ifrit:marker
        name="items"
        interface="ifrit.tests.IMarker"
     </ifrit:traversal_adapter>

    o Create a traversal adapter factory, which stamps the returned node
      with the marker interface passed in the corresponding 'marker'.

    o Bind that factory into 'module' under 'factory_name'.

    o Register that factory as providing the default 'IPathTraverser' adapter
      for the 'for' interface.
    """
    module = GlobalObject(
        title=u"Target module",
        description=u"Module into which the generated adapter factory "
                    u"will be added.",
        required=True,
        )

    factory_name = PythonIdentifier(
        title=u"Factory name",
        description=u"Name of the generated traversal adapter factory.",
        required=True,
        )

    for_ = Tokens(
        title=u"Specifications to be adapted",
        description=u"This should be a list of interfaces or classes",
        value_type=GlobalObject(missing_value=object()),
        required=False,
        )

class ITraversalMarkerDirective(Interface):
    """ Sub-directive schema for <ifrit:marker>.
    """
    name = PythonIdentifier(
        title=u"Traversal name",
        description=u"Name traversed by the adapter.",
        required=True,
        )

    interface = GlobalInterface(
        title=u"Merker Interface",
        description=u"Interface to be stamped onto returned nodes",
        required=False,
        )

def createTraversalAdapter(module, factory_name, for_, markers):
    """ Create an adapter factory using 'path'.
 
    o Seat the new schema into 'module' under 'factory_name'.
    """
    factory = ElementTraversalAdapterFactory(markers)
    setattr(module, factory_name, factory)
    provideAdapter(factory, for_, IPublishTraverse)
    provideAdapter(factory, for_ + (IRequest,), IPublishTraverse)

class TraversalAdapterDirective(object):

    def __init__(self, _context, module, factory_name, for_=None):
        # Directive handler for <ifrit:traversal_adapter> directive.
        self._context = _context
        self.module = module
        self.factory_name = factory_name
        self.for_ = for_
        self._markers = {}

    def marker(self, context, name, interface):
        self._markers[name] = interface

    def __call__(self):

        for_ = tuple(self.for_)
        self._context.action(
            discriminator=('adapter', for_, IPublishTraverse, ''),
            callable=createTraversalAdapter,
            args=(self.module, self.factory_name, for_, self._markers),
            )


